'''
synbiochem (c) University of Manchester 2016

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
import json
import urllib
import urllib2


def get_taxonomy_id(name):
    '''Gets the NCBI Taxonomy id from supplied name.'''
    url = 'http://www.ebi.ac.uk/ols/api/search?ontology=ncbitaxon' + \
        '&exact=true&queryFields=label&q=' + urllib.quote(name)

    response = urllib2.urlopen(url)
    data = json.loads(response.read())

    if data['response']['numFound'] == 1:
        term = data['response']['docs'][0]['obo_id']
        return term[term.find(':') + 1:]

    raise RuntimeError(name + ' not found in NCBI Taxonomy')


def search(term, exact=False):
    '''Gets the NCBI Taxonomy id from supplied name.'''
    url = 'http://www.ebi.ac.uk/ols/api/search?ontology=ncbitaxon' + \
        '&exact=' + str(exact) + '&queryFields=label&q=' + urllib.quote(term)

    response = urllib2.urlopen(url)
    data = json.loads(response.read())

    return [{'id': doc['obo_id'].replace('NCBITaxon:', ''),
             'name': doc['label']}
            for doc in data['response']['docs']]



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  pablocarbonell / neilswainston / alanwilliams
'''
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
import codecs
import json
import tempfile
import traceback
from xml.etree.ElementTree import ParseError

from synbiochem.utils import dna_utils, net_utils, sbol_utils


_DEFAULT_ID_PREFIX = 'SBC'
_SESSION_KEY = 'X-ICE-Authentication-SessionId'


class ICEEntry(object):
    '''Class to represent an ICE entry.'''

    def __init__(self, dna=None, typ=None, metadata=None):

        assert typ is not None or metadata is not None

        self.__dna = dna
        self.__dna_updated = dna is not None

        if metadata is None:
            self.__metadata = {'type': typ}
        else:
            if 'type' not in metadata:
                metadata['type'] = typ

            self.__metadata = metadata

    def get_ice_number(self):
        '''Gets the ICE number.'''
        return self.__metadata['id'] if 'id' in self.__metadata else None

    def get_ice_id(self):
        '''Gets the ICE id.'''
        ice_number = self.get_ice_number()
        return get_ice_id(ice_number) if ice_number is not None else None

    def get_record_id(self):
        '''Gets the ICE record id.'''
        return self.__metadata['recordId'] if 'recordId' in self.__metadata \
            else None

    def get_type(self):
        '''Gets the ICE type.'''
        return self.__metadata['type']

    def get_name(self):
        '''Gets the ICE name.'''
        return self.__metadata['name'] if 'name' in self.__metadata else ''

    def set_parameter(self, key, value):
        '''Sets a parameter.'''
        parameters = self.get_parameters()
        parameters[key] = value

        self.set_value('parameters',
                       [{'name': key, 'value': value}
                        for key, value in parameters.iteritems()])

    def get_parameters(self):
        '''Gets parameters.'''
        if 'parameters' in self.__metadata:
            return {parameter['name']: parameter['value']
                    for parameter in self.__metadata['parameters']}

        return {}

    def get_parameter(self, key):
        '''Gets parameter.'''
        parameters = self.get_parameters()
        return parameters.get(key, None)

    def get_metadata(self):
        '''Gets the metadata.'''
        return self.__metadata

    def get_dna(self):
        '''Gets the DNA object.'''
        return self.__dna

    def get_seq(self):
        '''Gets the sequence.'''
        return self.__dna['seq'] if self.__dna else ''

    def get_dna_updated(self):
        '''Gets the DNA object updated flag.'''
        return self.__dna_updated

    def set_values(self, new_metadata):
        '''Sets multiple metadata values.'''
        self.__metadata.update(new_metadata)

    def set_value(self, key, value):
        '''Sets a metadata value.'''
        self.__metadata[key] = value

    def set_dna(self, dna):
        '''Sets the DNA object.'''
        self.__dna_updated = self.__dna is not None \
            or dna is not None
        self.__dna = dna

    def unset_dna_updated(self):
        '''Sets the DNA object updated flag.'''
        self.__dna_updated = False

    def __repr__(self):
        return str(self.__metadata) + \
            ('\n' + str(self.__dna)
             if self.__dna is not None
             else '')


class ICEClient(object):
    '''Class representing an ICE client.'''

    def __init__(self, url, username, psswrd, id_prefix=_DEFAULT_ID_PREFIX,
                 group_names=None):
        self.__url = url[:-1] if url[-1] == '/' else url
        self.__username = username
        self.__psswrd = psswrd
        self.__id_prefix = id_prefix

        if group_names is None:
            group_names = []

        self.__headers = {'Accept': 'application/json',
                          'Content-Type': 'application/json'}

        self.__sid, self.__user, self.__email = self.reconnect()
        self.__headers[_SESSION_KEY] = self.__sid

        self.__group_ids = [group_id
                            for name, group_id
                            in self.get_groups().iteritems()
                            if name in group_names]

    def reconnect(self):
        '''Reconnects to ICE server.'''
        try:
            resp = self.__get_access_token(
                '/accesstoken', self.__username, self.__psswrd)
        except net_utils.NetworkError:
            resp = self.__get_access_token(
                '/accesstokens', self.__username, self.__psswrd)

        return resp['sessionId'], \
            resp['firstName'] + ' ' + resp['lastName'], \
            resp['email']

    def get_ice_entry(self, ice_id):
        '''Gets an ICEEntry object from the ICE database.'''
        metadata = self.__get_meta_data(ice_id)

        try:
            dna = self.__get_dna(ice_id) if metadata['hasSequence'] \
                else None
        except ParseError:
            # Take no action
            dna = None

        return ICEEntry(dna, metadata=metadata)

    def set_ice_entry(self, ice_entry):
        '''Saves an ICEEntry object in the ICE database.'''
        if ice_entry.get_ice_number() is None:
            response = self.__create_entry(ice_entry)
        else:
            response = self.__update_entry(ice_entry.get_ice_number(),
                                           self.__form_metadata(ice_entry))

        metadata = self.__get_meta_data(self.__get_ice_id(response['id']))
        ice_entry.set_values(metadata)

        if ice_entry.get_dna_updated():

            if 'hasSequence' in metadata and metadata['hasSequence']:
                self.__delete_seq(ice_entry.get_ice_number())

            dna = ice_entry.get_dna()

            if dna is not None:
                self.__upload_dna(ice_entry.get_record_id(),
                                  ice_entry.get_type(),
                                  dna)

                ice_entry.unset_dna_updated()

        metadata = self.__get_meta_data(self.__get_ice_id(response['id']))
        ice_entry.set_values(metadata)

        for group_id in self.__group_ids:
            self.add_permission(ice_entry.get_ice_id(), group_id)

        return response['id']

    def do_blast(self, seq, max_num=1024):
        '''Performs BLAST search against database.'''
        data = {'blastQuery': {'blastProgram': 'BLAST_N',
                               'sequence': seq.lower()},
                'parameters': {'retrieveCount': max_num}}
        return _read_resp(net_utils.post(self.__url + '/rest/search',
                                         json.dumps(data), self.__headers))

    def get_ice_entries_by_seq(self, seq):
        '''Returns entries matching supplied sequence.'''
        entries = []

        try:
            response = self.do_blast(seq)

            if 'results' in response:
                for result in response['results']:
                    if result['queryLength'] == int(result['alignment']):
                        entry = self.get_ice_entry(result['entryInfo']['id'])

                        if entry.get_seq() == seq:
                            entries.append(entry)
        except net_utils.NetworkError:
            # Sometime BLAST fails inexplicably (on ICE side).
            # Report error, return empty entries.
            traceback.print_exc()

        return entries

    def add_link(self, parent_id, child_id):
        '''Adds a link between a parent and child.'''
        url = self.__url + '/rest/parts/' + get_ice_number(parent_id) + \
            '/links'
        data = {'id': get_ice_number(child_id)}
        return _read_resp(net_utils.post(url,
                                         data=json.dumps(data),
                                         headers=self.__headers))

    def get_groups(self):
        '''Gets a group name to id map.'''
        url = self.__url + '/rest/groups?offset=0&limit=4096'
        groups = _read_resp(net_utils.get(url, headers=self.__headers))
        return {group['label']: group['id'] for group in groups['data']}

    def search_groups(self, term):
        '''Gets groups from search terms.'''
        url = self.__url + '/rest/groups/autocomplete?token=' + term
        return _read_resp(net_utils.get(url, headers=self.__headers))

    def add_permission(self, ice_id, group_id, read=True):
        '''Adds user permissions to a given ICE entry.'''
        url = self.__url + '/rest/parts/' + self.__get_ice_number(ice_id) + \
            '/permissions'
        data = {'type': 'READ_ENTRY' if read else 'WRITE_ENTRY',
                'article': 'GROUP',
                'articleId': group_id}

        return _read_resp(net_utils.post(url, json.dumps(data),
                                         self.__headers))

    def search(self, term, limit=5):
        '''Searches ICE.'''
        url = self.__url + '/rest/search?offset=0&limit=' + str(limit) + \
            '&sort=relevance&q="' + term + '"'

        return _read_resp(net_utils.get(url, self.__headers))

    def advanced_search(self, term, entry_type, limit=5):
        '''Searches ICE.'''
        data = {
            'queryString': term,
            'entryTypes': [entry_type],
            'parameters': {
                'sortField': 'RELEVANCE',
                'retrieveCount': limit,
            }
        }

        return _read_resp(net_utils.post(self.__url + '/rest/search',
                                         json.dumps(data),
                                         self.__headers))

    def __get_access_token(self, service, username, psswrd):
        '''Gets access token response.'''
        return _read_resp(net_utils.post(self.__url + '/rest' + service,
                                         json.dumps({'email': username,
                                                     'password': psswrd}),
                                         self.__headers))

    def __form_metadata(self, ice_entry):
        '''Forms metadata dictionary.'''
        metadata = ice_entry.get_metadata()

        if 'creator' not in metadata:
            metadata['creator'] = self.__user

        if 'creatorEmail' not in metadata:
            metadata['creatorEmail'] = self.__email

        return metadata

    def __upload_seq_file(self, record_id, typ, filename):
        '''Uploads a sequence file (not necessarily SBOL).'''
        return _read_resp(net_utils.post_file(self.__url +
                                              '/rest/file/sequence',
                                              {'file': open(filename, 'r'),
                                               'entryType': typ,
                                               'entryRecordId': record_id},
                                              {_SESSION_KEY: self.__sid}))

    def __delete_seq(self, ice_number):
        '''Deletes the sequence associated with supplied record id.'''
        net_utils.delete(self.__url + '/rest/parts/' + str(ice_number) +
                         '/sequence',
                         headers={_SESSION_KEY: self.__sid})

    def __get_meta_data(self, ice_id):
        '''Returns an ICE entry metadata.'''
        return _read_resp(net_utils.get(
            self.__url + '/rest/parts/' + self.__get_ice_number(ice_id),
            self.__headers))

    def __get_dna(self, ice_id):
        '''Gets the sequence ICE entry.'''
        url = self.__url + '/rest/file/' + self.__get_ice_number(ice_id) + \
            '/sequence/sbol1?sid=' + self.__sid
        temp_file = tempfile.NamedTemporaryFile(delete=False)

        with codecs.open(temp_file.name, 'w', 'utf-8') as text_file:
            text_file.write(net_utils.get(url))

        return sbol_utils.read(temp_file.name)

    def __create_entry(self, ice_entry):
        '''Creates a new ICE entry in the database.'''
        url = self.__url + '/rest/parts'
        return _read_resp(
            net_utils.post(url, json.dumps(self.__form_metadata(ice_entry)),
                           self.__headers))

    def __update_entry(self, ice_id, metadata):
        '''Updates an ICE entry in the database.'''
        ice_number = self.__get_ice_number(ice_id)
        url = self.__url + '/rest/parts/' + str(ice_number)

        return _read_resp(net_utils.put(url, json.dumps(metadata),
                                        self.__headers))

    def __upload_dna(self, record_id, typ, dna):
        '''Uploads an SBOLDocument to ICE database.'''
        sbol_file = tempfile.NamedTemporaryFile(suffix='.xml')
        sbol_utils.write(dna, sbol_file.name)
        return self.__upload_seq_file(record_id, typ, sbol_file.name)

    def __get_ice_number(self, ice_identifier):
        '''Maps ICE number to ICE id, i.e. from SBC000123 to 123,
        or if a number is supplied, returns the number.'''
        return get_ice_number(ice_identifier, self.__id_prefix)

    def __get_ice_id(self, ice_number):
        '''Maps ICE id to ICE number, i.e. from 123 to SBC000123.'''
        return get_ice_id(ice_number, self.__id_prefix)


class DNAWriter(object):
    '''Class for writing DNA objects to ICE.'''

    def __init__(self, ice_client):
        self.__ice_client = ice_client
        self.__cache = {}

    def submit(self, dna):
        '''Submits DNA to ICE (or checks for existing entry.'''
        if dna['seq'] in self.__cache:
            return self.__cache[dna['seq']]

        try:
            ice_entry = self.__ice_client.get_ice_entry(dna['disp_id'])
            ice_id, typ = ice_entry.get_ice_id(), ice_entry.get_type()
        except ValueError:
            # If disp_id is not a valid ICE id, try BLAST:
            ice_entries = self.__ice_client.get_ice_entries_by_seq(dna['seq'])

            if ice_entries:
                ice_entry = ice_entries[0]
                ice_id, typ = ice_entry.get_ice_id(), ice_entry.get_type()
            else:
                ice_id, typ = self.__write(dna)

        self.__cache[dna['seq']] = (ice_id, typ)
        return ice_id, typ

    def __write(self, dna):
        '''Writes DNA document to ICE.'''
        typ = 'PLASMID' if dna.get('typ', None) == dna_utils.SO_PLASMID \
            else 'PART'

        dna['parameters']['Full name'] = dna['name']

        ice_entry = ICEEntry(dna, typ)

        ice_entry.set_value('name', dna['name'][:125])
        ice_entry.set_value('shortDescription', dna['desc'])

        _add_params(ice_entry, dna)
        links = set(dna['links'])

        if dna.get('typ', None):
            links.add(dna['typ'])

        for feature in dna['features']:
            _add_params(ice_entry, feature)
            links |= set(feature['links'])

        ice_entry.set_value('links', list(links))

        entry_id = self.__ice_client.set_ice_entry(ice_entry)

        for child in dna['children']:
            par_ice_entry, _ = self.submit(child)
            self.__ice_client.add_link(entry_id, par_ice_entry)

        return entry_id, ice_entry.get_type()


def get_ice_number(ice_identifier, id_prefix=_DEFAULT_ID_PREFIX):
    '''Maps ICE number to ICE id, i.e. from SBC000123 to 123,
    or if a number is supplied, returns the number.'''
    try:
        ice_number = int(ice_identifier.replace(id_prefix, ''))
    except AttributeError:
        # "Ask forgiveness, not permission" and assume ice_identifier is
        # the ice_number:
        ice_number = ice_identifier

    return str(ice_number)


def get_ice_id(ice_number, id_prefix=_DEFAULT_ID_PREFIX):
    '''Maps ICE id to ICE number, i.e. from 123 to SBC000123.'''
    return id_prefix + format(ice_number, '06')


def _read_resp(response):
    '''Parses a string response into json.'''
    return json.loads(unicode(response)) if response else None


def _add_params(ice_entry, dna):
    '''Adds parameter values to ICEENtry.'''
    for key, value in dna['parameters'].iteritems():
        ice_entry.set_parameter(key, (', '.join([str(val) for val in value])
                                      if isinstance(value, list)
                                      else value))



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-public-methods
import random
import unittest

from synbiochem.utils import seq_utils


class CodonOptimiserTest(unittest.TestCase):
    '''Test class for CodonOptimiser.'''

    def test_get_codon_prob(self):
        '''Tests get_codon_prob method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertAlmostEquals(0.46, cod_opt.get_codon_prob('CTG'), 2)

    def test_get_codon_optim_seq(self):
        '''Tests get_codon_optim_seq method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        aa_codes = seq_utils.AA_CODES
        aa_codes.pop('Stop')

        aa_seq = ''.join([random.choice(aa_codes.values())
                          for _ in range(random.randint(100, 2500))])

        max_repeat_nuc = 5
        restr_enzyms = ['BsaI']
        dna_seq = cod_opt.get_codon_optim_seq(aa_seq,
                                              max_repeat_nuc=max_repeat_nuc,
                                              restr_enzyms=restr_enzyms)

        self.assertFalse(seq_utils.is_invalid(dna_seq, max_repeat_nuc,
                                              restr_enzyms))

    def test_get_random_codon(self):
        '''Tests get_random_codon method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertEquals('CTA', cod_opt.get_random_codon('L', ['CTG', 'TTA',
                                                                'CTT', 'TTG',
                                                                'CTC']))

    def test_get_random_codon_fail(self):
        '''Tests get_random_codon method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertRaises(
            ValueError, cod_opt.get_random_codon, 'M', ['ATG'])


class Test(unittest.TestCase):
    '''Test class for sequence_utils.'''

    def test_get_all_rev_trans(self):
        '''Tests get_all_rev_trans method.'''
        aa_seq = 'LS'
        rev_trans = seq_utils.get_all_rev_trans(aa_seq)
        self.assertEqual(len(rev_trans), 36)

    def test_get_codon_usage_organisms(self):
        '''Tests get_codon_usage_organisms method.'''
        organisms = seq_utils.get_codon_usage_organisms()
        expected = {'\'Arthromyces ramosus\'': '5451',
                    'thiosulfate-reducing anaerobe SRL 4198': '267367'}
        self.assertDictContainsSubset(expected, organisms)

    def test_find_invalid(self):
        '''Tests find_invalid method.'''
        seq = 'ggtctaaaaatttttttaaaaaccagagtttttt'

        self.assertEquals(seq_utils.find_invalid(seq, 5, ['BsaI']),
                          [10, 11, 28])

    def test_is_invalid(self):
        '''Tests is_invalid method.'''
        seq_inv = 'ggtctaaaaatttttttaaaaaccagagtttttt'
        self.assertTrue(seq_utils.is_invalid(seq_inv, 5, ['BsaI']))

        seq_val = 'ggtctaaaa'
        self.assertFalse(seq_utils.is_invalid(seq_val, 5, ['BsaI']))

    def test_get_random_dna(self):
        '''Tests get_random_dna method.'''
        lngth = random.randint(1000, 10000)
        self.assertEqual(lngth,
                         len(seq_utils.get_random_dna(lngth, 4, ['BsaI'])))

    def test_get_seq_by_melt_temp(self):
        '''Tests get_seq_by_melt_temp method.'''
        seq, _ = seq_utils.get_seq_by_melt_temp('agcgtgcgaagcgtgcgatcctcc', 70)
        self.assertEqual(seq, 'agcgtgcgaagcgtgcgatc')

    def test_get_rand_seq_by_melt_temp(self):
        '''Tests get_rand_seq_by_melt_temp method.'''
        target_temp = random.randint(50, 100)
        _, temp = seq_utils.get_rand_seq_by_melt_temp(target_temp, 4, ['BsaI'])
        self.assertTrue(abs(target_temp - temp) / target_temp < 0.025)

    def test_get_uniprot_values(self):
        '''Tests get_uniprot_values method.'''
        result = seq_utils.get_uniprot_values(['P19367', 'P42212'],
                                              ['organism-id',
                                               'protein names'], 1)

        expected = {'P19367': {'Entry': 'P19367',
                               'Organism ID': '9606',
                               'Protein names': ['Hexokinase-1',
                                                 'EC 2.7.1.1',
                                                 'Brain form hexokinase',
                                                 'Hexokinase type I',
                                                 'HK I']},
                    'P42212': {'Entry': 'P42212',
                               'Organism ID': '6100',
                               'Protein names': ['Green fluorescent protein']}}

        self.assertEquals(result, expected)

    def test_do_blast(self):
        '''Tests do_blast method.'''
        id_seq = {'test': seq_utils.get_random_dna(1024)}
        results = seq_utils.do_blast(id_seq, id_seq, evalue=10, word_size=4)

        alignments = []

        for result in results:
            for alignment in result.alignments:
                for hsp in alignment.hsps:
                    alignments.append(hsp)
                    print hsp

        self.assertGreater(len(alignments), 1)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=invalid-name
from igraph import Graph


def add_vertex(graph, name, kwds=None):
    '''Add vertex.'''
    try:
        return graph.vs.find(name)
    except ValueError:
        if not kwds:
            kwds = {}

        graph.add_vertex(name, **kwds)
        return graph.vs.find(name)


def add_edge(graph, vertex_from, vertex_to, kwds=None):
    '''Add edge.'''
    if not kwds:
        kwds = {}

    graph.add_edge(vertex_from.index, vertex_to.index, **kwds)


def get_roots(graph):
    '''Get roots.'''
    return [vs for vs, outdg in zip(graph.vs, graph.outdegree())
            if not outdg]



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=redundant-unittest-assert
# pylint: disable=too-many-public-methods
import getpass
import os
import unittest

from synbiochem.utils import ice_utils, sbol_utils
from synbiochem.utils.ice_utils import ICEClient, ICEEntry


class TestICEEntry(unittest.TestCase):
    '''Test class for ICEEntry.'''
    @classmethod
    def setUpClass(cls):
        ice_url = raw_input('ICE url: ')
        username = raw_input('ICE username: ')
        password = getpass.getpass(prompt='ICE password: ')
        cls.__ice_client = ICEClient(ice_url, username, password)

    def test_get_ice_number(self):
        '''Tests get_ice_number method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.__ice_client.set_ice_entry(ice_entry1)
        self.assertNotEqual(ice_entry1.get_ice_number(), None)

        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.__ice_client.set_ice_entry(ice_entry2)
        self.assertNotEqual(ice_entry2.get_ice_number(), None)

    def test_get_ice_number_none(self):
        '''Tests get_ice_number method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.assertEqual(ice_entry1.get_ice_number(), None)
        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.assertEqual(ice_entry2.get_ice_number(), None)

    def test_get_record_id(self):
        '''Tests get_record_id method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.__ice_client.set_ice_entry(ice_entry1)
        self.assertNotEqual(ice_entry1.get_record_id(), None)

        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.__ice_client.set_ice_entry(ice_entry2)
        self.assertNotEqual(ice_entry2.get_record_id(), None)

    def test_get_record_id_none(self):
        '''Tests get_record_id method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.assertEqual(ice_entry1.get_record_id(), None)
        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.assertEqual(ice_entry2.get_record_id(), None)

    def test_get_type(self):
        '''Tests get_type method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.assertEqual(ice_entry1.get_type(), 'PLASMID')
        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.assertEqual(ice_entry2.get_type(), 'PLASMID')

    def test_get_metadata(self):
        '''Tests get_metadata method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        self.assertEqual(ice_entry1.get_metadata()['type'], 'PLASMID')
        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        self.assertEqual(ice_entry2.get_metadata()['type'], 'PLASMID')

    def test_get_sbol_doc(self):
        '''Tests get_sbol_doc method.'''
        dna = _read('sbol.xml')
        ice_entry = ICEEntry(typ='PLASMID', dna=dna)
        self.assertEqual(ice_entry.get_dna(), dna)

    def test_set_value(self):
        '''Tests set_value method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        ice_entry1.set_value('creator', 'God')
        self.assertEqual(ice_entry1.get_metadata()['type'], 'PLASMID')
        self.assertEqual(ice_entry1.get_metadata()['creator'], 'God')
        self.__ice_client.set_ice_entry(ice_entry1)

        ice_entry2 = self.__ice_client.get_ice_entry(
            ice_entry1.get_ice_number())
        ice_entry2.set_value('creator', 'Aitor Karanka')
        self.__ice_client.set_ice_entry(ice_entry2)

        ice_entry3 = self.__ice_client.get_ice_entry(
            ice_entry1.get_ice_number())
        self.assertEqual(ice_entry3.get_metadata()['creator'], 'Aitor Karanka')

        ice_entry4 = ICEEntry(metadata={'type': 'PLASMID'})
        ice_entry4.set_value('creator', 'God')
        self.assertEqual(ice_entry4.get_metadata()['type'], 'PLASMID')
        self.assertEqual(ice_entry4.get_metadata()['creator'], 'God')

    def test_set_values(self):
        '''Tests set_values method.'''
        ice_entry1 = ICEEntry(typ='PLASMID')
        ice_entry1.set_values({'creator': 'God', 'name': 'Test'})
        self.assertEqual(ice_entry1.get_metadata()['type'], 'PLASMID')
        self.assertEqual(ice_entry1.get_metadata()['creator'], 'God')
        self.assertEqual(ice_entry1.get_metadata()['name'], 'Test')
        ice_entry2 = ICEEntry(metadata={'type': 'PLASMID'})
        ice_entry2.set_values({'creator': 'God', 'name': 'Test'})
        self.assertEqual(ice_entry2.get_metadata()['type'], 'PLASMID')
        self.assertEqual(ice_entry2.get_metadata()['creator'], 'God')
        self.assertEqual(ice_entry2.get_metadata()['name'], 'Test')

    def test_set_dna(self):
        '''Tests set_dna method.'''
        dna1 = _read('sbol.xml')
        dna2 = _read('sbol2.xml')

        ice_entry = ICEEntry(typ='PLASMID', dna=dna1)
        self.__ice_client.set_ice_entry(ice_entry)

        ice_entry.set_dna(dna2)
        self.__ice_client.set_ice_entry(ice_entry)

        self.assertEqual(ice_entry.get_seq(), dna2['seq'])

    def test_get_seq(self):
        '''Tests get_seq method.'''
        dna = _read('sbol.xml')

        ice_entry = ICEEntry(typ='PLASMID', dna=dna)
        self.__ice_client.set_ice_entry(ice_entry)
        ice_entry = self.__ice_client.get_ice_entry(ice_entry.get_ice_id())

        self.assertEqual(ice_entry.get_seq(), dna['seq'])

    def test_set_get_parameter(self):
        '''Tests set/get_parameter method.'''

        ice_entry = ICEEntry(typ='PLASMID')
        ice_entry.set_parameter('Cheese', 'Brie')
        self.__ice_client.set_ice_entry(ice_entry)
        ice_entry = self.__ice_client.get_ice_entry(ice_entry.get_ice_id())

        self.assertEqual(ice_entry.get_parameter('Cheese'), 'Brie')


class TestICEClient(unittest.TestCase):
    '''Test class for ICEClient.'''
    @classmethod
    def setUpClass(cls):
        ice_url = raw_input('ICE url: ')
        username = raw_input('ICE username: ')
        password = getpass.getpass(prompt='ICE password: ')
        cls.__ice_client = ICEClient(ice_url, username, password)

    def test_get_ice_entry(self):
        '''Tests get_ice_entry method.'''
        dna = _read('sbol.xml')

        ice_entry_in = ICEEntry(typ='PART', dna=dna)
        self.__ice_client.set_ice_entry(ice_entry_in)

        ice_entry_out = self.__ice_client.get_ice_entry(
            ice_entry_in.get_ice_number())
        self.assertEqual(ice_entry_out.get_seq(), dna['seq'])

    def test_set_ice_entry(self):
        '''Tests set_ice_entry method.'''
        dna_in = _read('sbol.xml')

        ice_entry_in = ICEEntry(typ='PLASMID', dna=dna_in)
        self.__ice_client.set_ice_entry(ice_entry_in)

        ice_entry_out = self.__ice_client.get_ice_entry(
            ice_entry_in.get_ice_number())

        self.assertNotEqual(ice_entry_out, None)

    def test_do_blast(self):
        '''Tests do_blast method.'''
        result = self.__ice_client.do_blast('tcgagaattcaaaagatctgagataggtaga' +
                                            'agctagacgagaaaccttcccaatcttatca' +
                                            'ttacgaaaggacgtccctatgagcctgatta')
        self.assertTrue(result['resultCount'] > 0)

    def test_get_ice_entries_by_seq(self):
        '''Tests get_ice_entries_by_seq method.'''
        dna = _read('sbol.xml')
        # dna.set_seq(sequence_utils.get_random_dna(4096))

        ice_entry = ICEEntry(typ='PLASMID', dna=dna)
        self.__ice_client.set_ice_entry(ice_entry)

        self.__ice_client.reconnect()
        result = self.__ice_client.get_ice_entries_by_seq(dna['seq'])
        self.assertTrue(len(result) > 0)

    def test_add_link(self):
        '''Tests add_link method.'''
        dna1 = _read('sbol.xml')
        dna2 = _read('sbol2.xml')

        ice_entry1 = ICEEntry(typ='PLASMID', dna=dna1)
        self.__ice_client.set_ice_entry(ice_entry1)

        ice_entry2 = ICEEntry(typ='PLASMID', dna=dna2)
        self.__ice_client.set_ice_entry(ice_entry2)

        self.__ice_client.add_link(
            ice_entry1.get_ice_number(), ice_entry2.get_ice_number())

        # "Refresh" (update metadata)
        ice_entry1 = self.__ice_client.get_ice_entry(ice_entry1.get_ice_id())
        ice_entry2 = self.__ice_client.get_ice_entry(ice_entry2.get_ice_id())

        self.assertTrue(ice_entry1.get_ice_number() in
                        [par['id']
                         for par in ice_entry2.get_metadata()['parents']])

        self.assertTrue(ice_entry2.get_ice_number() in
                        [link['id']
                         for link in ice_entry1.get_metadata()['linkedParts']])

    def test_search_groups(self):
        '''Tests get_group_id and search_groups method.'''
        groups = self.__ice_client.get_groups()

        for name in groups:
            groups = self.__ice_client.search_groups(name[:-1])
            self.assertTrue(name in [grp['label'] for grp in groups])

    def test_add_permission(self):
        '''Tests add_permission method.'''
        dna = _read('sbol.xml')
        ice_entry = ICEEntry(typ='PLASMID', dna=dna)
        self.__ice_client.set_ice_entry(ice_entry)

        groups = self.__ice_client.get_groups()

        for group_num in groups.values():
            self.__ice_client.add_permission(ice_entry.get_ice_id(), group_num)

        # If test progresses to here, it has succeeded:
        self.assertTrue(True)

    def test_search(self):
        '''Tests search method.'''
        resp = self.__ice_client.search('PLASMID')

        # If test progresses to here, it has succeeded:
        self.assertTrue(resp['resultCount'] > 0)

    def test_advanced_search(self):
        '''Tests search method.'''
        typ = 'PLASMID'
        resp = self.__ice_client.advanced_search('*', typ)

        self.assertTrue(len(resp['results']) == 5)

        for result in resp['results']:
            self.assertTrue(result['entryInfo']['type'] == typ)


class Test(unittest.TestCase):
    '''Test class for ice_utils.'''

    def test_get_ice_number(self):
        '''Tests get_ice_number method.'''
        self.assertEqual(ice_utils.get_ice_number('TEST00063', 'TEST'), '63')
        self.assertEqual(ice_utils.get_ice_number('TEST63', 'TEST'), '63')
        self.assertEqual(ice_utils.get_ice_number('63', 'TEST'), '63')
        self.assertEqual(ice_utils.get_ice_number(63, 'TEST'), '63')


def _read(filename):
    '''Reads sbol file.'''
    directory = os.path.dirname(os.path.realpath(__file__))
    return sbol_utils.read(os.path.join(directory, filename))


if __name__ == "__main__":
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
import re

from synbiochem.utils import seq_utils


class Mutation(object):
    '''Class to represent a mutation.'''

    def __init__(self, wt_res, pos, mut_res, typ='aa'):
        # Validate:
        alphabet = seq_utils.AA_CODES.values() if typ == 'aa' \
            else seq_utils.NUCLEOTIDES
        assert wt_res in alphabet
        assert isinstance(pos, (int, long)) and pos > 0
        assert mut_res in alphabet + ['-']  # Consider deletions

        self.__wt_res = wt_res
        self.__pos = pos
        self.__mut_res = mut_res

    def get_wt_res(self):
        '''Gets wt residue.'''
        return self.__wt_res

    def get_pos(self):
        '''Gets position.'''
        return self.__pos

    def get_mut_res(self):
        '''Gets mutation residue.'''
        return self.__mut_res

    def __repr__(self):
        return self.__wt_res + str(self.__pos) + self.__mut_res

    def __cmp__(self, other):
        pos_diff = self.__pos - other.get_pos()

        if pos_diff:
            return pos_diff

        return ord(self.__mut_res) - ord(other.get_mut_res())


def parse_mut_str(mut_str):
    '''Parse mutation string.'''
    return [Mutation(mut[0], int(mut[1]), mut[2])
            for mut in [re.compile(r'(\d*)').split(mutation)
                        for mutation in mut_str.split()]]


def get_mutations(wt_seq, mut_seq):
    '''Get Mutations.'''
    assert len(wt_seq) == len(mut_seq)

    mutations = []

    for (pos, aas) in enumerate(zip(wt_seq, mut_seq)):
        if aas[0] != aas[1]:
            mutations.append(Mutation(aas[0], pos + 1, aas[1]))

    return mutations


def apply_mutations(seq, mutations):
    '''Applies mutations to sequence.'''
    seq = list(seq)
    offset = 1

    for mutation in mutations:
        if mutation.get_wt_res() != seq[mutation.get_pos() - offset]:
            err = 'Invalid mutation at position %d. ' % mutation.get_pos() + \
                'Amino acid is %s ' % seq[mutation.get_pos() - offset] + \
                'but mutation is of %s.' % mutation.get_wt_res()

            raise ValueError(err)

        if mutation.get_mut_res() == '-':
            # Deletion:
            del seq[mutation.get_pos() - offset]
            offset += 1
        else:
            # Mutation:
            seq[mutation.get_pos() - offset] = mutation.get_mut_res()

    return ''.join(seq)



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-many-public-methods
import json
import os
import unittest

from Bio.Restriction import HgaI, MlyI
from synbiochem.utils import dna_utils, sbol_utils

import synbiochem.utils.test.test_sbol_utils as test_sbol_utils


class Test(unittest.TestCase):
    '''Test class for dna_utils.'''

    def test_copy(self):
        '''Tests copy method.'''
        directory = os.path.dirname(os.path.realpath(__file__))
        dna1 = sbol_utils.read(os.path.join(directory, 'sbol.xml'))
        dna2 = test_sbol_utils.round_trip(dna1.copy())
        self.assertEqual(dna1, dna2)

    def test_concat(self):
        '''Tests concat method.'''
        directory = os.path.dirname(os.path.realpath(__file__))
        dna2 = sbol_utils.read(os.path.join(directory, 'sbol2.xml'))
        dna3 = sbol_utils.read(os.path.join(directory, 'sbol3.xml'))
        concat_dna = test_sbol_utils.round_trip(dna_utils.concat([dna2, dna3]))

        self.assertFalse(concat_dna['features'][0]['forward'])

        self.assertEqual(len(dna2['features']) + len(dna3['features']),
                         len(concat_dna['features']))

    def test_json(self):
        '''Tests json roundtrip.'''
        directory = os.path.dirname(os.path.realpath(__file__))
        dna1 = sbol_utils.read(os.path.join(directory, 'sbol.xml'))
        params = json.loads(json.dumps(dna1))
        dna2 = dna_utils.DNA(**params)
        self.assertEqual(dna1, dna2)

    def test_app_restrict_site_linear(self):
        '''Tests apply_restriction_site method.'''
        _, dnas = _get_apply_restrict_site_dnas(MlyI, False)
        self.assertEquals([len(dna['seq']) for dna in dnas], [25, 831, 25])

    def test_app_restrict_site_circular(self):
        '''Tests apply_restriction_site method.'''
        _, dnas = _get_apply_restrict_site_dnas('MlyI', True)
        self.assertEquals([len(dna['seq']) for dna in dnas], [50, 831])

    def test_app_restrict_site_nomatch(self):
        '''Tests aplly_restriction_site method.'''
        parent, dnas = _get_apply_restrict_site_dnas(HgaI, False)
        self.assertEquals(len(dnas), 1)
        self.assertEquals(parent, dnas[0])


def _get_apply_restrict_site_dnas(restr, circ):
    '''Tests apply_restriction_site method.'''
    directory = os.path.dirname(os.path.realpath(__file__))
    par = sbol_utils.read(os.path.join(directory, 'restrict.xml'))
    return par, [test_sbol_utils.round_trip(dna)
                 for dna in dna_utils.apply_restricts(par, [restr], circ)]


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
from collections import defaultdict
import itertools
import operator
import os
import random
import re
from subprocess import call
import tempfile
import urllib
import urllib2

from Bio import SeqIO, SeqRecord
from Bio.Blast import NCBIXML
from Bio.Data import CodonTable
from Bio.Restriction import Restriction, Restriction_Dictionary
from Bio.Seq import Seq
from Bio.SeqUtils.MeltingTemp import Tm_NN
from requests.exceptions import ConnectionError
from synbiochem.biochem4j import taxonomy
from synbiochem.utils import thread_utils


NUCLEOTIDES = ['A', 'C', 'G', 'T']

AA_CODES = {'Ala': 'A',
            'Cys': 'C',
            'Asp': 'D',
            'Glu': 'E',
            'Phe': 'F',
            'Gly': 'G',
            'His': 'H',
            'Ile': 'I',
            'Lys': 'K',
            'Leu': 'L',
            'Met': 'M',
            'Asn': 'N',
            'Pro': 'P',
            'Gln': 'Q',
            'Arg': 'R',
            'Ser': 'S',
            'Thr': 'T',
            'Val': 'V',
            'Trp': 'W',
            'Tyr': 'Y',
            'Stop': '*'}

CODONS = {'A': [['G', 'C', 'ACGT']],
          'C': [['T', 'G', 'CT']],
          'D': [['G', 'A', 'CT']],
          'E': [['G', 'A', 'AG']],
          'F': [['T', 'T', 'CT']],
          'G': [['G', 'G', 'ACGT']],
          'H': [['C', 'A', 'CT']],
          'I': [['A', 'T', 'ACT']],
          'K': [['A', 'A', 'AG']],
          'L': [['C', 'T', 'ACGT'], ['T', 'T', 'AG']],
          'M': [['A', 'T', 'G']],
          'N': [['A', 'A', 'CT']],
          'P': [['C', 'C', 'ACGT']],
          'Q': [['C', 'A', 'AG']],
          'R': [['C', 'G', 'ACGT'], ['A', 'G', 'AG']],
          'S': [['T', 'C', 'ACGT'], ['A', 'G', 'CT']],
          'T': [['A', 'C', 'ACGT']],
          'V': [['G', 'T', 'ACGT']],
          'W': [['T', 'G', 'G']],
          'Y': [['T', 'A', 'CT']],
          'Stop': [['T', 'A', 'AG'], ['T', 'G', 'A']]}

NUCL_CODES = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'AG': 'R',
    'CT': 'Y',
    'CG': 'S',
    'AT': 'W',
    'GT': 'K',
    'AC': 'M',
    'CGT': 'B',
    'AGT': 'D',
    'ACT': 'H',
    'ACG': 'V',
    'ACGT': 'N',
}

INV_NUCL_CODES = {val: key for key, val in NUCL_CODES.items()}

NA = 'NA'
K = 'K'
TRIS = 'TRIS'
MG = 'MG'
DNTP = 'DNTP'

__DEFAULT_REAG_CONC = {NA: 0.05, K: 0, TRIS: 0, MG: 0.01, DNTP: 0}

AA_COD = defaultdict(list)

for cod, am_ac in \
        CodonTable.unambiguous_dna_by_name[
            'Standard'].forward_table.iteritems():
    AA_COD[am_ac].append(cod)


def get_codon_usage_organisms(expand=False, verbose=False):
    '''Gets name to taxonomy id dictionary of available codon usage tables.'''
    destination = os.path.join(os.path.expanduser('~'), 'codon')
    filename = 'expand.txt' if expand else 'normal.txt'
    filepath = os.path.join(destination, filename)

    if not os.path.exists(filepath):
        # Download:
        if not os.path.exists(destination):
            os.makedirs(destination)

        url = 'ftp://ftp.kazusa.or.jp/pub/codon/current/species.table'
        tmp = tempfile.NamedTemporaryFile(delete=False)
        urllib.urlretrieve(url, tmp.name)

        # Read:
        codon_orgs = _read_codon_usage_orgs_file(tmp.name)

        # Expand:
        if expand:
            _expand_codon_usage_orgs(codon_orgs, verbose)

        # Save:
        _write_codon_usage_orgs_file(codon_orgs, filepath)

        return codon_orgs
    else:
        return _read_codon_usage_orgs_file(filepath)


class CodonOptimiser(object):
    '''Class to support codon optimisation.'''

    def __init__(self, taxonomy_id):
        self.__taxonomy_id = taxonomy_id
        self.__aa_to_codon_prob = self.__get_codon_usage()
        self.__codon_prob = {item[0]: item[1]
                             for lst in self.__aa_to_codon_prob.values()
                             for item in lst}

        self.__codon_to_w = {}

        for key in self.__aa_to_codon_prob:
            aa_dict = dict([(a, b / self.__aa_to_codon_prob[key][0][1])
                            for a, b in self.__aa_to_codon_prob[key]])
            self.__codon_to_w.update(aa_dict)

    def get_codon_prob(self, codon):
        '''Gets the codon probability.'''
        return self.__codon_prob[codon]

    def get_codon_optim_seq(self, protein_seq, excl_codons=None,
                            max_repeat_nuc=float('inf'), restr_enzyms=None,
                            max_attempts=1000, tolerant=False, stepback=3):
        '''Returns a codon optimised DNA sequence.'''
        if max_repeat_nuc == float('inf') and restr_enzyms is None:
            return ''.join([self.get_random_codon(aa, excl_codons)
                            for aa in protein_seq])
        else:
            attempts = 0
            seq = ''
            i = 0
            blockage_i = -1
            inv_patterns = 0

            while attempts < max_attempts:
                amino_acid = protein_seq[i]
                new_seq = seq + self.get_random_codon(amino_acid, excl_codons)

                invalids = find_invalid(new_seq, max_repeat_nuc,
                                        restr_enzyms)

                if len(invalids) == inv_patterns or \
                        (attempts == max_attempts - 1 and tolerant):

                    if i == blockage_i:
                        if attempts == max_attempts - 1:
                            inv_patterns = inv_patterns + 1

                        attempts = 0

                    seq = new_seq

                    if i == len(protein_seq) - 1:
                        return seq

                    i += 1
                else:
                    blockage_i = max(i, blockage_i)
                    i = max(0, (invalids[-1] / 3) - stepback)
                    seq = seq[:i * 3]
                    attempts += 1

            raise ValueError('Unable to generate codon-optimised sequence.')

    def get_cai(self, dna_seq):
        '''Gets the CAI for a given DNA sequence.'''
        cai = 0

        for i in range(0, len(dna_seq), 3):
            cai += self.__codon_to_w[dna_seq[i:i + 3]]

        return cai / (len(dna_seq) / 3)

    def mutate(self, protein_seq, dna_seq, mutation_rate):
        '''Mutate a protein-encoding DNA sequence according to a
        supplied mutation rate.'''
        return ''.join([self.get_random_codon(amino_acid)
                        if random.random() < mutation_rate
                        else dna_seq[3 * i:3 * (i + 1)]
                        for i, amino_acid in enumerate(protein_seq)])

    def get_all_codons(self, amino_acid):
        '''Returns all codons for a given amino acid.'''
        return [t[0] for t in self.__aa_to_codon_prob[amino_acid]]

    def get_random_codon(self, amino_acid, excl_codons=None):
        '''Returns a random codon for a given amino acid,
        based on codon probability from the codon usage table.'''
        if excl_codons is None:
            excl_codons = []

        codon_usage = [codon_usage
                       for codon_usage in self.__aa_to_codon_prob[amino_acid]
                       if codon_usage[0] not in excl_codons]

        if not codon_usage:
            raise ValueError('No codons available for ' + amino_acid +
                             ' after excluding ' + str(excl_codons))

        while True:
            rand = random.random()
            cumulative_prob = 0

            for codon, prob in iter(reversed(codon_usage)):
                cumulative_prob += prob

                if cumulative_prob > rand:
                    return codon

    def __get_codon_usage(self):
        '''Gets the codon usage table for a given taxonomy id.'''
        aa_to_codon_prob = {aa_code: {} for aa_code in AA_CODES.values()}

        url = 'http://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=' \
            + self.__taxonomy_id + '&aa=1&style=GCG'

        in_codons = False

        for line in urllib2.urlopen(url):
            if line == '<PRE>\n':
                in_codons = True
            elif line == '</PRE>\n':
                break
            elif in_codons:
                values = re.split('\\s+', line)
                am_acid = 'Stop' if values[0] == 'End' else values[0]

                if am_acid in AA_CODES:
                    codon_prob = aa_to_codon_prob[AA_CODES[am_acid]]
                    codon_prob[values[1]] = float(values[3])

        aa_to_codon_prob.update((x, _scale(y))
                                for x, y in aa_to_codon_prob.items())

        return aa_to_codon_prob


def find_invalid(seq, max_repeat_nuc=float('inf'), restr_enzyms=None):
    '''Finds invalid sequences.'''
    inv = []
    seq = seq.upper()

    # Invalid repeating nucleotides:
    if max_repeat_nuc != float('inf'):
        pattern = [''.join([nucl] * (max_repeat_nuc + 1))
                   for nucl in NUCLEOTIDES]
        pattern = re.compile(r'(?=(' + '|'.join(pattern) + '))')

        inv = [m.start() for m in pattern.finditer(seq)]

    # Invalid restriction sites:
    if restr_enzyms:
        for rest_enz in [_get_restr_type(name) for name in restr_enzyms]:
            inv.extend(rest_enz.search(Seq(seq)))

    return inv


def is_invalid(seq, max_repeat_nuc=float('inf'), restr_enzyms=None):
    '''Checks whether a sequence is valid.'''
    return len(find_invalid(seq, max_repeat_nuc, restr_enzyms)) > 0


def get_all_rev_trans(aa_seq):
    '''Returns all reverse translations of amino acid sequence.'''
    codons = [AA_COD[aa] for aa in aa_seq.strip()]
    return [''.join(t) for t in list(itertools.product(*codons))]


def get_random_dna(length, max_repeat_nuc=float('inf'), restr_enzyms=None):
    '''Returns a random sequence of DNA of the supplied length,
    while adhering to a maximum number of repeating nucleotides.'''
    max_attempts = 100
    attempts = 0
    len_add = 16

    seq = ''

    while True:
        attempts += 1

        if attempts > max_attempts:
            raise ValueError('Unable to optimise sequence.')

        while len(seq) < length:
            seq += _get_random_dna(len_add)

            if is_invalid(seq, max_repeat_nuc, restr_enzyms):
                seq = seq[:-len_add]

        if not is_invalid(seq, max_repeat_nuc, restr_enzyms):
            return seq[:length]

    return None


def mutate_seq(seq, mutations=1, alphabet=None):
    '''Mutates sequence.'''
    if alphabet is None:
        alphabet = NUCLEOTIDES

    seq_new = seq

    for _ in range(mutations):
        move = random.random()
        pos = int(random.random() * len(seq))
        base = random.choice(alphabet)

        # Insert:
        if move < 0.1:
            seq_new = seq_new[1:pos + 1] + base + seq_new[pos + 1:]

        # Delete:
        elif move < 0.2:
            seq_new = base + seq_new[:pos] + seq_new[pos + 1:]

        # Replace:
        else:
            seq_new = seq_new[:pos] + base + seq_new[pos + 1:]

    return seq_new


def get_melting_temp(dna1, dna2=None, reag_concs=None, strict=True):
    '''Calculates melting temperarure of DNA sequence against its
    complement, or against second DNA sequence using Nearest-Neighbour
    method.'''
    assert len(dna1) > 1

    reagent_concs = __DEFAULT_REAG_CONC

    if reag_concs is not None:
        reagent_concs.update(reag_concs)

    reagent_conc = {k: v * 1000 for k, v in reagent_concs.iteritems()}
    dnac1 = 30

    return Tm_NN(dna1, check=True, strict=strict, c_seq=dna2, shift=0,
                 Na=reagent_conc[NA], K=reagent_conc[K],
                 Tris=reagent_conc[TRIS], Mg=reagent_conc[MG],
                 dNTPs=reagent_conc[DNTP],
                 dnac1=dnac1, dnac2=dnac1, selfcomp=dna2 is None,
                 saltcorr=7)


def get_seq_by_melt_temp(seq, target_melt_temp, forward=True,
                         terminii=None,
                         reagent_concs=None,
                         tol=0.025):
    '''Returns a subsequence close to desired melting temperature.'''
    if terminii is None:
        terminii = ['A', 'C', 'G', 'T']
    else:
        terminii = [term.upper() for term in terminii]

    best_delta_tm = float('inf')
    best_subseq = ''
    best_melt_temp = float('NaN')
    in_tol = False

    for i in range(1, len(seq)):
        subseq = seq[:(i + 1)] if forward else seq[-(i + 1):]
        melt_temp = get_melting_temp(subseq, None, reagent_concs)

        if subseq[-1 if forward else 0].upper() in terminii:
            delta_tm = abs(melt_temp - target_melt_temp)

            if delta_tm / target_melt_temp < tol:
                in_tol = True

                if delta_tm < best_delta_tm:
                    best_delta_tm = delta_tm
                    best_subseq = subseq
                    best_melt_temp = melt_temp
            elif in_tol:
                return best_subseq, best_melt_temp

    raise ValueError('Unable to get sequence of required melting temperature')


def get_rand_seq_by_melt_temp(target_melt_temp,
                              max_repeat_nuc=float('inf'),
                              restr_enzyms=None,
                              reagent_concs=None,
                              tol=0.025):
    '''Returns a random close to desired melting temperature.'''
    seq = random.choice(NUCLEOTIDES)

    while True:
        seq += random.choice(NUCLEOTIDES)

        if is_invalid(seq, max_repeat_nuc, restr_enzyms):
            seq = seq[:-random.choice(range(len(seq)))]
            continue

        melt_temp = get_melting_temp(seq, None, reagent_concs)

        delta_tm = abs(melt_temp - target_melt_temp)

        if delta_tm / target_melt_temp < tol:
            return seq, melt_temp

    raise ValueError('Unable to get sequence of required melting temperature')


def get_uniprot_values(uniprot_ids, fields, batch_size=128, verbose=False,
                       num_threads=0):
    '''Gets dictionary of ids to values from Uniprot.'''
    values = []

    if num_threads:
        thread_pool = thread_utils.ThreadPool(num_threads)

        for i in xrange(0, len(uniprot_ids), batch_size):
            thread_pool.add_task(_get_uniprot_batch, uniprot_ids, i,
                                 batch_size, fields, values, verbose)

        thread_pool.wait_completion()
    else:
        for i in xrange(0, len(uniprot_ids), batch_size):
            _get_uniprot_batch(uniprot_ids, i, batch_size, fields, values,
                               verbose)

    return {value['Entry']: value for value in values}


def search_uniprot(query, fields, limit=128):
    '''Gets dictionary of ids to values from Uniprot.'''
    values = []

    url = 'http://www.uniprot.org/uniprot/?query=' + urllib.quote(query) + \
        '&sort=score&limit=' + str(limit) + \
        '&format=tab&columns=id,' + ','.join([urllib.quote(field)
                                              for field in fields])

    _parse_uniprot_data(url, values)

    return values


def do_blast(id_seqs_subjects, id_seqs_queries, program='blastn',
             dbtype='nucl', evalue=1.0, word_size=28):
    '''Performs BLAST of query sequences against subject sequences.'''
    db_filename = write_fasta(id_seqs_subjects)
    query_filename = write_fasta(id_seqs_queries)
    result_file = tempfile.NamedTemporaryFile(prefix='blast_result_',
                                              suffix='.xml',
                                              delete=False)
    log_file = tempfile.NamedTemporaryFile(prefix='makeblastdb_log',
                                           suffix='.txt',
                                           delete=False)

    call(['makeblastdb',
          '-in', db_filename,
          '-out', db_filename,
          '-dbtype', dbtype,
          '-logfile', log_file.name])

    call([program,
          '-query', query_filename,
          '-db', db_filename,
          '-out', result_file.name,
          '-evalue', str(evalue),
          '-word_size', str(word_size),
          '-outfmt', '5'])

    return NCBIXML.parse(open(result_file.name))


def do_clustal(id_seqs):
    '''Performs Clustal Omega multiple sequence alignment.'''
    in_file = write_fasta(id_seqs)
    result_file = tempfile.NamedTemporaryFile(prefix='clustalo_result_',
                                              suffix='.fasta',
                                              delete=False)

    call(['clustalo',
          '-i', in_file,
          '-o', result_file.name,
          '--force'])

    return read_fasta(result_file.name)


def read_fasta(filename):
    '''Reads a fasta file.'''
    with open(filename, 'rU') as fle:
        seqs = {record.id: str(record.seq)
                for record in SeqIO.parse(fle, 'fasta')}

    return seqs


def write_fasta(id_seqs, filename=None):
    '''Writes a fasta file.'''
    if filename is None:
        temp_file = tempfile.NamedTemporaryFile(prefix='fasta_', suffix='.txt',
                                                delete=False)
        filename = temp_file.name

    records = [SeqRecord.SeqRecord(Seq(seq), str(seq_id), '', '')
               for seq_id, seq in id_seqs.iteritems()]

    SeqIO.write(records, filename, 'fasta')

    return filename


def _scale(codon_usage):
    '''Scale codon usage values to add to 1.'''
    codon_usage = dict([(key, value / sum(codon_usage.values()))
                        for key, value in codon_usage.items()])

    return sorted(codon_usage.items(), key=operator.itemgetter(1),
                  reverse=True)


def _get_random_dna(length):
    '''Returns a random sequence of DNA of the supplied length.'''
    return ''.join(random.choice(['A', 'C', 'G', 'T']) for _ in range(length))


def _get_uniprot_batch(uniprot_ids, i, batch_size, fields, values, verbose):
    '''Get batch of Uniprot data.'''
    if verbose:
        print 'seq_utils: getting Uniprot values ' + str(i) + ' - ' + \
            str(min(i + batch_size, len(uniprot_ids))) + ' / ' + \
            str(len(uniprot_ids))

    batch = uniprot_ids[i:min(i + batch_size, len(uniprot_ids))]
    query = '+or+'.join(['id:' + uniprot_id for uniprot_id in batch])
    url = 'http://www.uniprot.org/uniprot/?query=' + query + \
        '&format=tab&columns=id,' + ','.join([urllib.quote(field)
                                              for field in fields])

    _parse_uniprot_data(url, values)


def _parse_uniprot_data(url, values):
    '''Parses Uniprot data.'''
    headers = None

    try:
        for line in urllib2.urlopen(url):
            tokens = line.strip().split('\t')

            if headers is None:
                headers = tokens
            else:
                resp = dict(zip(headers, tokens))

                if 'Protein names' in resp:
                    regexp = re.compile(r'(?<=\()[^)]*(?=\))|^[^(][^()]*')
                    names = regexp.findall(resp.pop('Protein names'))
                    resp['Protein names'] = [nme.strip() for nme in names]

                for key in resp:
                    if key.startswith('Cross-reference'):
                        resp[key] = resp[key].split(';')

                values.append(resp)
    except urllib2.HTTPError, err:
        print err


def _read_codon_usage_orgs_file(filename):
    '''Reads Codon Usage Database table of species file.'''
    codon_orgs = {}

    with open(filename, 'r') as textfile:
        next(textfile)

        for line in textfile:
            tokens = line.strip().split('\t')
            codon_orgs[tokens[0]] = tokens[1]

    return codon_orgs


def _expand_codon_usage_orgs(codon_orgs, verbose, max_errors=16):
    '''Expand Codon Usage Db table of species with children and synonyms.'''
    for tax_id in codon_orgs.values():

        if verbose:
            print 'Expanding codon usage for NCBI Taxonomy id: ' + tax_id

        errors = 0
        success = False

        while not success:
            try:
                for name in taxonomy.get_synonyms_by_id(tax_id):
                    _add_codon_usage_org(codon_orgs, name, tax_id)

                for child in taxonomy.get_children_by_id(tax_id):
                    _add_codon_usage_org(codon_orgs, child['name'], tax_id)

                    for name in child['names']:
                        _add_codon_usage_org(codon_orgs, name, tax_id)

                success = True

            except ConnectionError, err:
                errors += 1

                if errors == max_errors:
                    raise err


def _add_codon_usage_org(codon_orgs, name, tax_id):
    '''Adds name to codon_orgs.'''
    if name not in codon_orgs:
        codon_orgs[name] = tax_id


def _write_codon_usage_orgs_file(codon_orgs, filepath):
    '''Writes Codon Usage Database table of species file.'''
    with open(filepath, 'w+') as fle:
        fle.write('Name\tId\n')

        for name, tax_id in codon_orgs.iteritems():
            fle.write(name + '\t' + tax_id + '\n')


def _get_restr_type(name):
    '''Gets RestrictionType from name.'''
    types = [
        x for _, (x, y) in Restriction_Dictionary.typedict.items()
        if name in y][0]

    enz_types = tuple(getattr(Restriction, typ)
                      for typ in types)

    return Restriction.RestrictionType(
        str(name), enz_types, Restriction_Dictionary.rest_dict[name])



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
import re


def pairwise(iterable):
    '''s -> (s0,s1), (s1,s2), (s2, s3), ...'''
    return [(iterable[i], iterable[i + 1]) for i in range(len(iterable) - 1)]


def sort(sortable):
    ''''Sort the given iterable in the way that humans expect.'''
    def convert(text):
        '''Convert char to int if digit.'''
        return int(text) if text.isdigit() else text

    def alphanum_key(key):
        '''Generate alphanumeric key.'''
        return [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(sortable, key=alphanum_key)



'''
synbiochem (c) University of Manchester 2017

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
import os
import re
import shutil
import subprocess

import numpy as np
import pandas as pd


def create_db(neo4j_root_loc, nodes_files, rels_files, db_name='graph.db',
              ignore_duplicates=False, delimiter=';', array_delimiter='|'):
    '''Creates the database from csv files.

    Import format is:

    neo4j-admin import --nodes NODES_FILE --relationships RELS_FILE

    '''
    # Apply types to import files:
    for filename in nodes_files + rels_files:
        _type_file(filename, delimiter)

    # Stop database:
    bin_loc = os.path.join(neo4j_root_loc, 'bin')
    subprocess.call([os.path.join(bin_loc, 'neo4j'), 'stop'])

    # Remove database directory if it already exists:
    db_loc = os.path.join(neo4j_root_loc, 'data', 'databases', db_name)

    if os.path.exists(db_loc):
        shutil.rmtree(db_loc)

    # Import database:
    admin_loc = os.path.join(bin_loc, 'neo4j-admin')

    params = [admin_loc,
              'import',
              '--database', db_name]

    for node_file in nodes_files:
        params.extend(['--nodes', node_file])

    for rels_file in rels_files:
        params.extend(['--relationships', rels_file])

    if ignore_duplicates:
        params.append('--ignore-duplicate-nodes')

    # not sure why, but this bit is required by ENI model
    # seems like one more bug regarding reading csv
    params.extend(['--multiline-fields', 'True'])

    params.extend(['--delimiter', delimiter])
    params.extend(['--array-delimiter', array_delimiter])

    subprocess.call(params)


def _type_file(filename, sep):
    '''Update column names in file based on dtype.'''
    df = pd.read_csv(filename, sep=sep, low_memory=False)
    _type_cols(df)
    df.to_csv(filename, sep, index=False)


def _type_cols(df):
    '''Update column names based on dtype.'''
    # Id columns regexps:
    id_str = r'(:ID|:START_ID|:END_ID)'
    id_regex = re.compile(id_str)

    # Ensure absence of colons in non-id column names:
    type_cols = {col: col.replace(':', '_')
                 for col in df
                 if col != ':TYPE' and col != ':LABEL' and ':' in col
                 and not id_regex.search(col)}

    # Convert appropriate cols to boolean:
    for col in df:
        if True in pd.unique(df[col]) and df[col].dtype is np.dtype('O'):
            type_cols.update({col: col + ':boolean'})
            df[col] = df[col].astype(str).str.lower()

    # Add types for int and float columns *except* those of ids,
    # which are always assumed to be string:
    col_grouped_by_type = df.columns.to_series().groupby(df.dtypes).groups
    col_grouped_by_type.pop(np.dtype('O'))

    for type_name, cols in col_grouped_by_type.iteritems():
        type_name = ''.join([i for i in str(type_name).lower()
                             if not i.isdigit()])

        type_cols.update({col: col + ':' + type_name
                          for col in cols
                          if not id_regex.search(col)})

    df.rename(columns=type_cols, inplace=True)
